import json
import logging
import os
import sys
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path

HERE = Path(__file__).parent.resolve()
PROJECT = HERE.parent

sys.path.insert(0, str(PROJECT / "src"))  # <-- Use development version of library
logging.basicConfig(level=logging.DEBUG)

from validate_pyproject import caching, http  # noqa: E402

SCHEMA_STORE = "https://json.schemastore.org/pyproject.json"


def iter_test_urls():
    with caching.as_file(http.open_url, SCHEMA_STORE) as f:
        store = json.load(f)
        for _, tool in store["properties"]["tool"]["properties"].items():
            if "$ref" in tool and tool["$ref"].startswith(("http://", "https://")):
                yield tool["$ref"]

    files = PROJECT.glob("**/test_config.json")
    for file in files:
        content = json.loads(file.read_text("utf-8"))
        for _, url in content.get("tools", {}).items():
            if url.startswith(("http://", "https://")):
                yield url


def download(url):
    return caching.as_file(http.open_url, url).close()
    # ^-- side-effect only: write cached file


def download_all(cache: str):
    with ThreadPoolExecutor(max_workers=5) as executor:
        return list(executor.map(download, set(iter_test_urls())))  # Consume iterator


if __name__ == "__main__":
    cache = os.getenv("VALIDATE_PYPROJECT_CACHE_REMOTE")
    if not cache:
        raise SystemExit("Please define VALIDATE_PYPROJECT_CACHE_REMOTE")

    Path(cache).mkdir(parents=True, exist_ok=True)
    downloads = download_all(cache)
    assert len(downloads) > 0, f"empty {downloads=!r}"  # noqa
    files = list(map(print, Path(cache).iterdir()))
    assert len(files) > 0, f"empty {files=!r}"  # noqa



# The code in this module is based on a similar code from `ini2toml` (originally
# published under the MPL-2.0 license)
# https://github.com/abravalheri/ini2toml/blob/49897590a9254646434b7341225932e54f9626a3/LICENSE.txt

# ruff: noqa: C408
# Unnecessary `dict` call (rewrite as a literal)

import argparse
import io
import json
import logging
import sys
from contextlib import contextmanager
from itertools import chain
from textwrap import dedent, wrap
from typing import (
    Callable,
    Dict,
    Generator,
    Iterator,
    List,
    NamedTuple,
    Sequence,
    Tuple,
    Type,
    TypeVar,
)

from . import __version__
from . import _tomllib as tomllib
from .api import Validator
from .errors import ValidationError
from .plugins import PluginProtocol, PluginWrapper
from .plugins import list_from_entry_points as list_plugins_from_entry_points
from .remote import RemotePlugin, load_store

_logger = logging.getLogger(__package__)
T = TypeVar("T", bound=NamedTuple)

_REGULAR_EXCEPTIONS = (ValidationError, tomllib.TOMLDecodeError)


@contextmanager
def critical_logging() -> Generator[None, None, None]:
    """Make sure the logging level is set even before parsing the CLI args"""
    try:
        yield
    except Exception:  # pragma: no cover
        if "-vv" in sys.argv or "--very-verbose" in sys.argv:
            setup_logging(logging.DEBUG)
        raise


_STDIN = argparse.FileType("r")("-")

META: Dict[str, dict] = {
    "version": dict(
        flags=("-V", "--version"),
        action="version",
        version=f"{__package__} {__version__}",
    ),
    "input_file": dict(
        dest="input_file",
        nargs="*",
        # default=[_STDIN],  # postponed to facilitate testing
        type=argparse.FileType("r"),
        help="TOML file to be verified (`stdin` by default)",
    ),
    "enable": dict(
        flags=("-E", "--enable-plugins"),
        nargs="+",
        default=(),
        dest="enable",
        metavar="PLUGINS",
        help="Enable ONLY the given plugins (ALL plugins are enabled by default).",
    ),
    "disable": dict(
        flags=("-D", "--disable-plugins"),
        nargs="+",
        dest="disable",
        default=(),
        metavar="PLUGINS",
        help="Enable ALL plugins, EXCEPT the ones given.",
    ),
    "verbose": dict(
        flags=("-v", "--verbose"),
        dest="loglevel",
        action="store_const",
        const=logging.INFO,
        help="set logging level to INFO",
    ),
    "very_verbose": dict(
        flags=("-vv", "--very-verbose"),
        dest="loglevel",
        action="store_const",
        const=logging.DEBUG,
        help="set logging level to DEBUG",
    ),
    "dump_json": dict(
        flags=("--dump-json",),
        action="store_true",
        help="Print the JSON equivalent to the given TOML",
    ),
    "tool": dict(
        flags=("-t", "--tool"),
        action="append",
        dest="tool",
        help="External tools file/url(s) to load, of the form name=URL#path",
    ),
    "store": dict(
        flags=("--store",),
        help="Load a pyproject.json file and read all the $ref's into tools "
        "(see https://json.schemastore.org/pyproject.json)",
    ),
}


class CliParams(NamedTuple):
    input_file: List[io.TextIOBase]
    plugins: List[PluginWrapper]
    tool: List[str]
    store: str
    loglevel: int = logging.WARNING
    dump_json: bool = False


def __meta__(plugins: Sequence[PluginProtocol]) -> Dict[str, dict]:
    """'Hyper parameters' to instruct :mod:`argparse` how to create the CLI"""
    meta = {k: v.copy() for k, v in META.items()}
    meta["enable"]["choices"] = {p.tool for p in plugins}
    meta["input_file"]["default"] = [_STDIN]  # lazily defined to facilitate testing
    return meta


@critical_logging()
def parse_args(
    args: Sequence[str],
    plugins: Sequence[PluginProtocol],
    description: str = "Validate a given TOML file",
    get_parser_spec: Callable[[Sequence[PluginProtocol]], Dict[str, dict]] = __meta__,
    params_class: Type[T] = CliParams,  # type: ignore[assignment]
) -> T:
    """Parse command line parameters

    Args:
      args: command line parameters as list of strings (for example  ``["--help"]``).

    Returns: command line parameters namespace
    """
    epilog = ""
    if plugins:
        epilog = f"The following plugins are available:\n\n{plugins_help(plugins)}"

    parser = argparse.ArgumentParser(
        description=description, epilog=epilog, formatter_class=Formatter
    )
    for cli_opts in get_parser_spec(plugins).values():
        parser.add_argument(*cli_opts.pop("flags", ()), **cli_opts)

    parser.set_defaults(loglevel=logging.WARNING)
    params = vars(parser.parse_args(args))
    enabled = params.pop("enable", ())
    disabled = params.pop("disable", ())
    params["tool"] = params["tool"] or []
    params["store"] = params["store"] or ""
    params["plugins"] = select_plugins(plugins, enabled, disabled)
    return params_class(**params)  # type: ignore[call-overload, no-any-return]


Plugins = TypeVar("Plugins", bound=PluginProtocol)


def select_plugins(
    plugins: Sequence[Plugins],
    enabled: Sequence[str] = (),
    disabled: Sequence[str] = (),
) -> List[Plugins]:
    available = list(plugins)
    if enabled:
        available = [p for p in available if p.tool in enabled]
    if disabled:
        available = [p for p in available if p.tool not in disabled]
    return available


def setup_logging(loglevel: int) -> None:
    """Setup basic logging

    Args:
      loglevel: minimum loglevel for emitting messages
    """
    logformat = "[%(levelname)s] %(message)s"
    logging.basicConfig(level=loglevel, stream=sys.stderr, format=logformat)


@contextmanager
def exceptions2exit() -> Generator[None, None, None]:
    try:
        yield
    except _ExceptionGroup as group:
        for prefix, ex in group:
            print(prefix)
            _logger.error(str(ex) + "\n")
        raise SystemExit(1) from None
    except _REGULAR_EXCEPTIONS as ex:
        _logger.error(str(ex))
        raise SystemExit(1) from None
    except Exception as ex:  # pragma: no cover
        _logger.error(f"{ex.__class__.__name__}: {ex}\n")
        _logger.debug("Please check the following information:", exc_info=True)
        raise SystemExit(1) from None


def run(args: Sequence[str] = ()) -> int:
    """Wrapper allowing :obj:`Translator` to be called in a CLI fashion.

    Instead of returning the value from :func:`Translator.translate`, it prints the
    result to the given ``output_file`` or ``stdout``.

    Args:
      args (List[str]): command line parameters as list of strings
          (for example  ``["--verbose", "setup.cfg"]``).
    """
    args = args or sys.argv[1:]
    plugins = list_plugins_from_entry_points()
    params: CliParams = parse_args(args, plugins)
    setup_logging(params.loglevel)
    tool_plugins = [RemotePlugin.from_str(t) for t in params.tool]
    if params.store:
        tool_plugins.extend(load_store(params.store))
    validator = Validator(params.plugins, extra_plugins=tool_plugins)

    exceptions = _ExceptionGroup()
    for file in params.input_file:
        try:
            _run_on_file(validator, params, file)
        except _REGULAR_EXCEPTIONS as ex:
            exceptions.add(f"Invalid {_format_file(file)}", ex)

    exceptions.raise_if_any()

    return 0


def _run_on_file(validator: Validator, params: CliParams, file: io.TextIOBase) -> None:
    if file in (sys.stdin, _STDIN):
        print("Expecting input via `stdin`...", file=sys.stderr, flush=True)

    toml_equivalent = tomllib.loads(file.read())
    validator(toml_equivalent)
    if params.dump_json:
        print(json.dumps(toml_equivalent, indent=2))
    else:
        print(f"Valid {_format_file(file)}")


main = exceptions2exit()(run)


class Formatter(argparse.RawTextHelpFormatter):
    # Since the stdlib does not specify what is the signature we need to implement in
    # order to create our own formatter, we are left no choice other then overwrite a
    # "private" method considered to be an implementation detail.

    def _split_lines(self, text: str, width: int) -> List[str]:
        return list(chain.from_iterable(wrap(x, width) for x in text.splitlines()))


def plugins_help(plugins: Sequence[PluginProtocol]) -> str:
    return "\n".join(_format_plugin_help(p) for p in plugins)


def _flatten_str(text: str) -> str:
    text = " ".join(x.strip() for x in dedent(text).splitlines()).strip()
    text = text.rstrip(".,;").strip()
    return (text[0].lower() + text[1:]).strip()


def _format_plugin_help(plugin: PluginProtocol) -> str:
    help_text = plugin.help_text
    help_text = f": {_flatten_str(help_text)}" if help_text else ""
    return f"* {plugin.tool!r}{help_text}"


def _format_file(file: io.TextIOBase) -> str:
    if hasattr(file, "name") and file.name:
        return f"file: {file.name}"
    return "file"  # pragma: no cover


class _ExceptionGroup(Exception):
    _members: List[Tuple[str, Exception]]

    def __init__(self) -> None:
        self._members = []
        super().__init__()

    def add(self, prefix: str, ex: Exception) -> None:
        self._members.append((prefix, ex))

    def __iter__(self) -> Iterator[Tuple[str, Exception]]:
        return iter(self._members)

    def raise_if_any(self) -> None:
        number = len(self._members)
        if number == 1:
            print(self._members[0][0])
            raise self._members[0][1]
        if number > 0:
            raise self



# This file is execfile()d with the current directory set to its containing dir.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os
import sys

# -- Path setup --------------------------------------------------------------

__location__ = os.path.dirname(__file__)
sys.path.insert(0, __location__)
sys.path.insert(0, os.path.join(__location__, "../src"))

# -- Dynamically generated docs ----------------------------------------------
import _gendocs

output_dir = os.path.join(__location__, "api")
module_dir = os.path.join(__location__, "../src/validate_pyproject")
_gendocs.gen_stubs(module_dir, output_dir)

# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.ifconfig",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.extlinks",
    "sphinx_copybutton",
    "sphinxemoji.sphinxemoji",
    "sphinx-jsonschema",
    "sphinxarg.ext",
]

# ----------------------------------
# JSON Schema settings
jsonschema_options = {
    "lift_title": True,
    "lift_description": True,
    "lift_definitions": True,
    "auto_reference": True,
    "auto_target": True,
}
# ----------------------------------

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

try:
    from validate_pyproject import __version__, dist_name
except ImportError:
    __version__, dist_name = "", "validate-pyproject"


# General information about the project.
project = dist_name
copyright = "2021, Anderson Bravalheri"
repository = "https://github.com/abravalheri/validate-pyproject"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# version: The short X.Y version.
# release: The full version, including alpha/beta/rc tags.
# If you don't need the separation provided between version and release,
# just set them both to the same value.
version = __version__

if not version or version.lower() == "unknown":
    version = os.getenv("READTHEDOCS_VERSION", "unknown")  # automatically set by RTD

release = version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", ".venv"]

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"
pygments_dark_style = "monokai"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False

# If this is True, todo emits a warning for each TODO entries. The default is False.
todo_emit_warnings = True


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
# html_theme = "alabaster"
html_theme = "furo"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    "navigation_with_keys": True,
    "light_css_variables": {
        "color-brand-primary": "#2980B9",
        "color-brand-content": "#005CA0",
        "color-brand-muted": "#E7F2FA",
        "color-brand-logo-background": "#156EAD",
    },
    "dark_css_variables": {
        "color-brand-content": "#0A93FB",
        "color-brand-muted": "#00091A",
    },
}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = project

# A shorter title for the navigation bar.  Default is the same as html_title.
html_short_title = project

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = ""

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_css_files = [
    "custom-adjustments.css",  # Avoid name clashes with the theme
]

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = "validate-pyproject-doc"


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ("letterpaper" or "a4paper").
    # "papersize": "letterpaper",
    # The font size ("10pt", "11pt" or "12pt").
    # "pointsize": "10pt",
    # Additional stuff for the LaTeX preamble.
    # "preamble": "",
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
    (
        "index",
        "user_guide.tex",
        "validate-pyproject Documentation",
        "Anderson Bravalheri",
        "manual",
    )
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = ""

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True

# -- External mapping --------------------------------------------------------
python_version = ".".join(map(str, sys.version_info[0:2]))
intersphinx_mapping = {
    "sphinx": ("https://www.sphinx-doc.org/en/master", None),
    "python": ("https://docs.python.org/" + python_version, None),
    "matplotlib": ("https://matplotlib.org", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "sklearn": ("https://scikit-learn.org/stable", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/reference", None),
    "setuptools": ("https://setuptools.pypa.io/en/stable/", None),
    "pyscaffold": ("https://pyscaffold.org/en/stable", None),
    "fastjsonschema": ("https://horejsek.github.io/python-fastjsonschema/", None),
    "pypa": ("https://packaging.python.org/en/latest/", None),
}
extlinks = {
    "issue": (f"{repository}/issues/%s", "issue #%s"),
    "pr": (f"{repository}/pull/%s", "PR #%s"),
    "discussion": (f"{repository}/discussions/%s", "discussion #%s"),
    "pypi": ("https://pypi.org/project/%s", "%s"),
    "github": ("https://github.com/%s", "%s"),
    "user": ("https://github.com/sponsors/%s", "@%s"),
}

print(f"loading configurations for {project} {version} ...", file=sys.stderr)



# ruff: noqa: C408
# Unnecessary `dict` call (rewrite as a literal)

import json
import logging
import sys
from functools import partial, wraps
from pathlib import Path
from types import MappingProxyType
from typing import Any, Dict, List, Mapping, NamedTuple, Sequence

from .. import cli
from ..plugins import PluginProtocol, PluginWrapper
from ..plugins import list_from_entry_points as list_plugins_from_entry_points
from ..remote import RemotePlugin, load_store
from . import pre_compile

if sys.platform == "win32":  # pragma: no cover
    from subprocess import list2cmdline as arg_join
else:  # pragma: no cover
    from shlex import join as arg_join


_logger = logging.getLogger(__package__)


def JSON_dict(name: str, value: str) -> Dict[str, Any]:
    try:
        return ensure_dict(name, json.loads(value))
    except json.JSONDecodeError as ex:
        raise ValueError(f"Invalid JSON: {value}") from ex


META: Dict[str, dict] = {
    "output_dir": dict(
        flags=("-O", "--output-dir"),
        default=".",
        type=Path,
        help="Path to the directory where the files for embedding will be generated "
        "(default: current working directory)",
    ),
    "main_file": dict(
        flags=("-M", "--main-file"),
        default="__init__.py",
        help="Name of the file that will contain the main `validate` function"
        "(default: `%(default)s`)",
    ),
    "replacements": dict(
        flags=("-R", "--replacements"),
        default="{}",
        type=wraps(JSON_dict)(partial(JSON_dict, "replacements")),
        help="JSON string (don't forget to quote) representing a map between strings "
        "that should be replaced in the generated files and their replacement, "
        "for example: \n"
        '-R \'{"from packaging import": "from .._vendor.packaging import"}\'',
    ),
    "tool": dict(
        flags=("-t", "--tool"),
        action="append",
        dest="tool",
        help="External tools file/url(s) to load, of the form name=URL#path",
    ),
    "store": dict(
        flags=("--store",),
        help="Load a pyproject.json file and read all the $ref's into tools "
        "(see https://json.schemastore.org/pyproject.json)",
    ),
}


def ensure_dict(name: str, value: Any) -> dict:
    if not isinstance(value, dict):
        msg = f"`{value.__class__.__name__}` given (value = {value!r})."
        raise ValueError(f"`{name}` should be a dict. {msg}")
    return value


class CliParams(NamedTuple):
    plugins: List[PluginWrapper]
    output_dir: Path = Path(".")
    main_file: str = "__init__.py"
    replacements: Mapping[str, str] = MappingProxyType({})
    loglevel: int = logging.WARNING
    tool: Sequence[str] = ()
    store: str = ""


def parser_spec(
    plugins: Sequence[PluginProtocol],
) -> Dict[str, dict]:
    common = ("version", "enable", "disable", "verbose", "very_verbose")
    cli_spec = cli.__meta__(plugins)
    meta = {k: v.copy() for k, v in META.items()}
    meta.update({k: cli_spec[k].copy() for k in common})
    return meta


def run(args: Sequence[str] = ()) -> int:
    args = args if args else sys.argv[1:]
    cmd = f"python -m {__package__} " + arg_join(args)
    plugins = list_plugins_from_entry_points()
    desc = 'Generate files for "pre-compiling" `validate-pyproject`'
    prms = cli.parse_args(args, plugins, desc, parser_spec, CliParams)
    cli.setup_logging(prms.loglevel)

    tool_plugins: List[PluginProtocol] = [RemotePlugin.from_str(t) for t in prms.tool]
    if prms.store:
        tool_plugins.extend(load_store(prms.store))

    pre_compile(
        prms.output_dir,
        prms.main_file,
        cmd,
        prms.plugins,
        prms.replacements,
        extra_plugins=tool_plugins,
    )
    return 0


main = cli.exceptions2exit()(run)


if __name__ == "__main__":
    main()



import logging
import os
from itertools import chain
from unittest.mock import Mock

import pytest

from validate_pyproject import api, formats

_chain_iter = chain.from_iterable

# The following examples were taken by inspecting some opensource projects in the python
# community
ENTRYPOINT_EXAMPLES = {
    "django": {
        "console_scripts": {
            "django-admin": "django.core.management:execute_from_command_line"
        }
    },
    "pandas": {
        "pandas_plotting_backends": {"matplotlib": "pandas:plotting._matplotlib"},
    },
    "PyScaffold": {
        "console_scripts": {"putup": "pyscaffold.cli:run"},
        "pyscaffold.cli": {
            "config": "pyscaffold.extensions.config:Config",
            "interactive": "pyscaffold.extensions.interactive:Interactive",
            "venv": "pyscaffold.extensions.venv:Venv",
            "namespace": "pyscaffold.extensions.namespace:Namespace",
            "no_skeleton": "pyscaffold.extensions.no_skeleton:NoSkeleton",
            "pre_commit": "pyscaffold.extensions.pre_commit:PreCommit",
            "no_tox": "pyscaffold.extensions.no_tox:NoTox",
            "gitlab": "pyscaffold.extensions.gitlab_ci:GitLab",
            "cirrus": "pyscaffold.extensions.cirrus:Cirrus",
            "no_pyproject": "pyscaffold.extensions.no_pyproject:NoPyProject",
        },
    },
    "setuptools-scm": {
        "distutils.setup_keywords": {
            "use_scm_version": "setuptools_scm.integration:version_keyword",
        },
        "setuptools.file_finders": {
            "setuptools_scm": "setuptools_scm.integration:find_files",
        },
        "setuptools.finalize_distribution_options": {
            "setuptools_scm": "setuptools_scm.integration:infer_version",
        },
        "setuptools_scm.files_command": {
            ".hg": "setuptools_scm.file_finder_hg:hg_find_files",
            ".git": "setuptools_scm.file_finder_git:git_find_files",
        },
        "setuptools_scm.local_scheme": {
            "node-and-date": "setuptools_scm.version:get_local_node_and_date",
            "node-and-timestamp": "setuptools_scm.version:get_local_node_and_timestamp",
            "dirty-tag": "setuptools_scm.version:get_local_dirty_tag",
            "no-local-version": "setuptools_scm.version:get_no_local_node",
        },
        "setuptools_scm.parse_scm": {
            ".hg": "setuptools_scm.hg:parse",
            ".git": "setuptools_scm.git:parse",
        },
        "setuptools_scm.parse_scm_fallback": {
            ".hg_archival.txt": "setuptools_scm.hg:parse_archival",
            "PKG-INFO": "setuptools_scm.hacks:parse_pkginfo",
            "pip-egg-info": "setuptools_scm.hacks:parse_pip_egg_info",
            "setup.py": "setuptools_scm.hacks:fallback_version",
        },
        "setuptools_scm.version_scheme": {
            "guess-next-dev": "setuptools_scm.version:guess_next_dev_version",
            "post-release": "setuptools_scm.version:postrelease_version",
            "python-simplified-semver": "setuptools_scm.version:simplified_semver_version",
            "release-branch-semver": "setuptools_scm.version:release_branch_semver_version",
            "no-guess-dev": "setuptools_scm.version:no_guess_dev_version",
            "calver-by-date": "setuptools_scm.version:calver_by_date",
        },
    },
    "anyio": {
        "pytest11": {
            "anyio": "anyio.pytest_plugin",
        },
    },
}


@pytest.mark.parametrize(
    "example", _chain_iter(v.keys() for v in ENTRYPOINT_EXAMPLES.values())
)
def test_entrypoint_group(example):
    assert formats.python_entrypoint_group(example)


@pytest.mark.parametrize(
    "example",
    _chain_iter(
        _chain_iter(e.keys() for e in v.values()) for v in ENTRYPOINT_EXAMPLES.values()
    ),
)
def test_entrypoint_name(example):
    assert formats.python_entrypoint_name(example)


@pytest.mark.parametrize("example", [" invalid", "=invalid", "[invalid]", "[invalid"])
def test_entrypoint_invalid_name(example):
    assert formats.python_entrypoint_name(example) is False


@pytest.mark.parametrize("example", ["val[id", "also valid"])
def test_entrypoint_name_not_recommended(example, caplog):
    caplog.set_level(logging.WARNING)
    assert formats.python_entrypoint_name(example) is True
    assert "does not follow recommended pattern" in caplog.text


@pytest.mark.parametrize(
    "example",
    _chain_iter(
        _chain_iter(e.values() for e in v.values())
        for v in ENTRYPOINT_EXAMPLES.values()
    ),
)
def test_entrypoint_references(example):
    assert formats.python_entrypoint_reference(example)
    assert formats.pep517_backend_reference(example)
    assert formats.pep517_backend_reference(example.replace(":", "."))


def test_entrypoint_references_with_extras():
    example = "test.module:func [invalid"
    assert formats.python_entrypoint_reference(example) is False

    example = "test.module:func [valid]"
    assert formats.python_entrypoint_reference(example)
    assert formats.pep517_backend_reference(example) is False

    example = "test.module:func [valid, extras]"
    assert formats.python_entrypoint_reference(example)

    example = "test.module:func [??inva#%@!lid??]"
    assert formats.python_entrypoint_reference(example) is False


@pytest.mark.parametrize("example", ["module", "invalid-module"])
def test_invalid_entrypoint_references(example):
    result = example == "module"
    assert formats.python_entrypoint_reference(example) is result


@pytest.mark.parametrize("example", ["λ", "a", "_"])
def test_valid_python_identifier(example):
    assert formats.python_identifier(example)


@pytest.mark.parametrize("example", ["a.b", "x+y", " a", "☺"])
def test_invalid_python_identifier(example):
    assert formats.python_identifier(example) is False


@pytest.mark.parametrize(
    "example",
    [
        "0.9.10",
        "1988.12",
        "1.01rc1",
        "0.99a9",
        "3.14b5",
        "1.42.post0",
        "1.73a2.post0",
        "2.23.post6.dev0",
        "3!6.0",
        "1.0+abc.7",
        "v4.0.1",
    ],
)
def test_valid_pep440(example):
    assert formats.pep440(example)


@pytest.mark.parametrize(
    "example",
    [
        "0-9-10",
        "v4.0.1.mysuffix",
        "p4.0.2",
    ],
)
def test_invalid_pep440(example):
    assert formats.pep440(example) is False


@pytest.mark.parametrize(
    "example",
    [
        "~= 0.9, >= 1.0, != 1.3.4.*, < 2.0",
        ">= 1.4.5, == 1.4.*",
        "~= 2.2.post3",
        "!= 1.1.post1",
    ],
)
def test_valid_pep508_versionspec(example):
    assert formats.pep508_versionspec(example)


@pytest.mark.parametrize(
    "example",
    [
        "~ 0.9, ~> 1.0, - 1.3.4.*",
        "- 1.3.4.*",
        "~> 1.0",
        "~ 0.9",
        "@ file:///localbuilds/pip-1.3.1.zip",
        'v1.0; python_version<"2.7"',
    ],
)
def test_invalid_pep508_versionspec(example):
    assert formats.pep508_versionspec(example) is False


@pytest.mark.parametrize(
    "example",
    [
        "https://python.org",
        "http://python.org",
        "http://localhost:8000",
        "ftp://python.org",
        "scheme://netloc/path;parameters?query#fragment",
    ],
)
def test_valid_url(example):
    assert formats.url(example)


@pytest.mark.parametrize(
    "example",
    [
        "",
        42,
        "p@python.org",
        "http:python.org",
        "/python.org",
    ],
)
def test_invalid_url(example):
    assert formats.url(example) is False


@pytest.mark.parametrize(
    "example",
    [
        "ab",
        "ab.c.d",
        "abc._d.λ",
    ],
)
def test_valid_module_name(example):
    assert formats.python_module_name(example) is True


@pytest.mark.parametrize(
    "example",
    [
        "-",
        " ",
        "ab-cd",
        ".example",
    ],
)
def test_invalid_module_name(example):
    assert formats.python_module_name(example) is False


@pytest.mark.parametrize(
    "example",
    [
        "pip-run",
        "abc-d-λ",
        "abc-d-λ.xyz-e",
        "abc-d.λ-xyz-e",
    ],
)
def test_valid_module_name_relaxed(example):
    assert formats.python_module_name_relaxed(example) is True


@pytest.mark.parametrize(
    "example",
    [
        "pip run",
        "-pip-run",
        "pip-run-",
        "pip-run-stubs",
    ],
)
def test_invalid_module_name_relaxed(example):
    assert formats.python_module_name_relaxed(example) is False


@pytest.mark.parametrize(
    "example",
    [
        "MIT",
        "Bsd-3-clause",
        "mit and (apache-2.0 or bsd-2-clause)",
        "MIT OR GPL-2.0-or-later OR (FSFUL AND BSD-2-Clause)",
        "GPL-3.0-only WITH Classpath-exception-2.0 OR BSD-3-Clause",
        "LicenseRef-Special-License OR CC0-1.0 OR Unlicense",
        "LicenseRef-Public-Domain",
        "licenseref-proprietary",
        "LicenseRef-Beerware-4.2",
        "(LicenseRef-Special-License OR LicenseRef-OtherLicense) OR Unlicense",
    ],
)
def test_valid_pep639_license_expression(example):
    assert formats.SPDX(example) is True


@pytest.mark.parametrize(
    "example",
    [
        "",
        "Use-it-after-midnight",
        "LicenseRef-License with spaces",
        "LicenseRef-License_with_underscores",
        "or",
        "and",
        "with",
        "mit or",
        "mit and",
        "mit with",
        "or mit",
        "and mit",
        "with mit",
        "(mit",
        "mit)",
        "mit or or apache-2.0",
        # Missing an operator before `(`.
        "mit or apache-2.0 (bsd-3-clause and MPL-2.0)",
        # "2-BSD-Clause is not a valid license.
        "Apache-2.0 OR 2-BSD-Clause",
    ],
)
def test_invalid_pep639_license_expression(example):
    assert formats.SPDX(example) is False


class TestClassifiers:
    """The ``_TroveClassifier`` class and ``_download_classifiers`` are part of the
    private API and therefore need to be tested.

    By constantly testing them we can make sure the URL used to download classifiers and
    the format they are presented are still supported by PyPI.

    If at any point these tests start to fail, we know that we need to change strategy.
    """

    VALID_CLASSIFIERS = (
        "Development Status :: 5 - Production/Stable",
        "Framework :: Django",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3 :: Only",
        "private :: not really a classifier",
    )

    def test_does_not_break_public_function_detection(self):
        # See https://github.com/abravalheri/validate-pyproject/issues/12

        # When `trove_classifiers` is defined from the dependency package
        # it will be a function.
        # When it is defined based on the download, it will be a custom object.

        # In both cases the `_TroveClassifier` class should not be made public,
        # but the instance should

        trove_classifier = formats._TroveClassifier()
        _formats = Mock(
            _TroveClassifier=formats._TroveClassifier,
            trove_classifiers=trove_classifier,
        )
        fns = api._get_public_functions(_formats)
        assert fns == {"trove-classifier": trove_classifier}

        # Make sure the object and the function have the same name
        assert "trove-classifier" in api.FORMAT_FUNCTIONS
        normalized_name = trove_classifier.__name__.replace("_", "-")
        assert normalized_name == "trove-classifier"
        assert normalized_name in api.FORMAT_FUNCTIONS

        func_name = trove_classifier.__name__
        assert getattr(formats, func_name) in api.FORMAT_FUNCTIONS.values()

    def test_download(self):
        try:
            classifiers = formats._download_classifiers()
        except Exception as ex:
            pytest.xfail(f"Error with download: {ex.__class__.__name__} - {ex}")
        assert isinstance(classifiers, str)
        assert bytes(classifiers, "utf-8")

    def test_downloaded(self, monkeypatch):
        if os.name != "posix":
            # Mock on Windows (problems with SSL)
            downloader = Mock(return_value="\n".join(self.VALID_CLASSIFIERS))
            monkeypatch.setattr(formats, "_download_classifiers", downloader)

        validator = formats._TroveClassifier()
        assert validator("Made Up :: Classifier") is False
        assert validator.downloaded is not None
        assert validator.downloaded is not False
        assert len(validator.downloaded) > 3

    def test_valid_download_only_once(self, monkeypatch):
        if os.name == "posix":
            # Really download to make sure the API is still exposed by PyPI
            downloader = Mock(side_effect=formats._download_classifiers)
        else:
            # Mock on Windows (problems with SSL)
            downloader = Mock(return_value="\n".join(self.VALID_CLASSIFIERS))

        monkeypatch.setattr(formats, "_download_classifiers", downloader)
        validator = formats._TroveClassifier()
        for classifier in self.VALID_CLASSIFIERS:
            assert validator(classifier) is True
        downloader.assert_called_once()

    @pytest.mark.parametrize(
        "no_network", ("NO_NETWORK", "VALIDATE_PYPROJECT_NO_NETWORK")
    )
    def test_always_valid_with_no_network(self, monkeypatch, no_network):
        monkeypatch.setenv(no_network, "1")
        validator = formats._TroveClassifier()
        assert validator("Made Up :: Classifier") is True
        assert not validator.downloaded
        assert validator("Other Made Up :: Classifier") is True
        assert not validator.downloaded

    def test_always_valid_with_skip_download(self):
        validator = formats._TroveClassifier()
        validator._disable_download()
        assert validator("Made Up :: Classifier") is True
        assert not validator.downloaded
        assert validator("Other Made Up :: Classifier") is True
        assert not validator.downloaded

    def test_always_valid_after_download_error(self, monkeypatch):
        def _failed_download():
            raise OSError()

        monkeypatch.setattr(formats, "_download_classifiers", _failed_download)
        validator = formats._TroveClassifier()
        assert validator("Made Up :: Classifier") is True
        assert not validator.downloaded
        assert validator("Other Made Up :: Classifier") is True
        assert not validator.downloaded


def test_private_classifier():
    assert formats.trove_classifier("private :: Keep Off PyPI") is True
    assert formats.trove_classifier("private:: Keep Off PyPI") is False



"""
conftest.py for validate_pyproject.

Read more about conftest.py under:
- https://docs.pytest.org/en/stable/fixture.html
- https://docs.pytest.org/en/stable/writing_plugins.html
"""

from pathlib import Path
from typing import List

import pytest

HERE = Path(__file__).parent.resolve()


def pytest_configure(config):
    config.addinivalue_line("markers", "uses_network: tests may try to download files")


def collect(base: Path) -> List[str]:
    return [str(f.relative_to(base)) for f in base.glob("**/*.toml")]


@pytest.fixture(params=collect(HERE / "examples"))
def example(request) -> Path:
    return HERE / "examples" / request.param


@pytest.fixture(params=collect(HERE / "invalid-examples"))
def invalid_example(request) -> Path:
    return HERE / "invalid-examples" / request.param


@pytest.fixture(params=collect(HERE / "remote/examples"))
def remote_example(request) -> Path:
    return HERE / "remote/examples" / request.param


@pytest.fixture(params=collect(HERE / "remote/invalid-examples"))
def remote_invalid_example(request) -> Path:
    return HERE / "remote/invalid-examples" / request.param



# The code in this module is mostly borrowed/adapted from PyScaffold and was originally
# published under the MIT license
# The original PyScaffold license can be found in 'NOTICE.txt'
from __future__ import annotations

import sys
from collections import defaultdict
from importlib.metadata import EntryPoint
from types import ModuleType
from typing import Callable, TypeVar

import pytest

from validate_pyproject import plugins
from validate_pyproject.plugins import ErrorLoadingPlugin, PluginWrapper, StoredPlugin

EXISTING = (
    "setuptools",
    "distutils",
)

T = TypeVar("T", bound=Callable)


def test_load_from_entry_point__error():
    # This module does not exist, so Python will have some trouble loading it
    # EntryPoint(name, value, group)
    entry = "mypkg.SOOOOO___fake___:activate"
    fake = EntryPoint("fake", entry, "validate_pyproject.tool_schema")
    with pytest.raises(ErrorLoadingPlugin):
        plugins.load_from_entry_point(fake)


def is_entry_point(ep):
    return all(hasattr(ep, attr) for attr in ("name", "load"))


def test_iterate_entry_points():
    plugin_iter = plugins.iterate_entry_points("validate_pyproject.tool_schema")
    assert hasattr(plugin_iter, "__iter__")
    pluging_list = list(plugin_iter)
    assert all(is_entry_point(e) for e in pluging_list)
    name_list = [e.name for e in pluging_list]
    for ext in EXISTING:
        assert ext in name_list


def test_list_from_entry_points():
    # Should return a list with all the plugins registered in the entrypoints
    pluging_list = plugins.list_from_entry_points()
    orig_len = len(pluging_list)
    plugin_names = " ".join(e.tool for e in pluging_list)
    for example in EXISTING:
        assert example in plugin_names

    # a filtering function can be passed to avoid loading plugins that are not needed
    pluging_list = plugins.list_from_entry_points(
        filtering=lambda e: e.name != "setuptools"
    )
    plugin_names = " ".join(e.tool for e in pluging_list)
    assert len(pluging_list) == orig_len - 1
    assert "setuptools" not in plugin_names


class TestPluginWrapper:
    def test_empty_help_text(self):
        def _fn1(_):
            return {}

        pw = plugins.PluginWrapper("name", _fn1)
        assert pw.help_text == ""

        def _fn2(_):
            """Help for `${tool}`"""
            return {}

        pw = plugins.PluginWrapper("name", _fn2)
        assert pw.help_text == "Help for `name`"


class TestStoredPlugin:
    def test_empty_help_text(self):
        def _fn1(_):
            return {}

        pw = plugins.StoredPlugin("name", {}, "id1", 0)
        assert pw.help_text == ""

        def _fn2(_):
            """Help for `${tool}`"""
            return {}

        pw = plugins.StoredPlugin("name", {"description": "Help for me"}, "id2", 0)
        assert pw.help_text == "Help for me"


class _FakeEntryPoints:
    def __init__(
        self,
        monkeypatch: pytest.MonkeyPatch,
        group: str = "__NOT_SPECIFIED__",
        data: defaultdict[str, list[EntryPoint]] | None = None,
    ):
        self._monkeypatch = monkeypatch
        self._group = group
        self._data = defaultdict(list) if data is None else data
        self.get = self._data.__getitem__

    def group(self, group: str) -> _FakeEntryPoints:
        return _FakeEntryPoints(self._monkeypatch, group, self._data)

    def reverse(self) -> _FakeEntryPoints:
        data = defaultdict(list, {k: list(reversed(v)) for k, v in self._data.items()})
        return _FakeEntryPoints(self._monkeypatch, self._group, data)

    def __call__(self, *, name: str, value: str) -> Callable[[T], T]:
        def fake_entry_point(impl: T) -> T:
            ep = EntryPoint(name=name, value=value, group=self._group)
            self._data[ep.group].append(ep)
            module, _, func = ep.value.partition(":")
            if module not in sys.modules:
                self._monkeypatch.setitem(sys.modules, module, ModuleType(module))
            self._monkeypatch.setattr(sys.modules[module], func, impl, raising=False)
            return impl

        return fake_entry_point


def test_multi_plugins(monkeypatch):
    fake_eps = _FakeEntryPoints(monkeypatch, group="validate_pyproject.multi_schema")
    fake_eps(name="f", value="test_module:f")(
        lambda: {
            "tools": {"example#frag": {"$id": "example1"}},
            "schemas": [
                {"$id": "example2"},
                {"$id": "example3"},
            ],
        }
    )
    monkeypatch.setattr(plugins, "iterate_entry_points", fake_eps.get)

    lst = plugins.list_from_entry_points()
    assert len(lst) == 3

    (fragmented,) = (e for e in lst if e.tool)
    assert fragmented.tool == "example"
    assert fragmented.fragment == "frag"
    assert fragmented.schema == {"$id": "example1"}


@pytest.mark.parametrize("epname", ["aaa", "zzz"])
def test_combined_plugins(monkeypatch, epname):
    fake_eps = _FakeEntryPoints(monkeypatch)
    multi_eps = fake_eps.group("validate_pyproject.multi_schema")
    tool_eps = fake_eps.group("validate_pyproject.tool_schema")
    multi_eps(name=epname, value="test_module:f")(
        lambda: {
            "tools": {
                "example1": {"$id": "example1"},
                "example2": {"$id": "example2"},
                "example3": {"$id": "example3"},
            }
        }
    )
    tool_eps(name="example1", value="test_module:f1")(lambda _: {"$id": "ztool1"})
    tool_eps(name="example2", value="test_module:f2")(lambda _: {"$id": "atool2"})
    tool_eps(name="example4", value="test_module:f2")(lambda _: {"$id": "tool4"})

    monkeypatch.setattr(plugins, "iterate_entry_points", fake_eps.get)

    lst = plugins.list_from_entry_points()
    print(lst)
    assert len(lst) == 4

    assert lst[0].tool == "example1"
    assert isinstance(lst[0], PluginWrapper)

    assert lst[1].tool == "example2"
    assert isinstance(lst[1], PluginWrapper)

    assert lst[2].tool == "example3"
    assert isinstance(lst[2], StoredPlugin)

    assert lst[3].tool == "example4"
    assert isinstance(lst[3], PluginWrapper)


def test_several_multi_plugins(monkeypatch):
    fake_eps = _FakeEntryPoints(monkeypatch, "validate_pyproject.multi_schema")
    fake_eps(name="zzz", value="test_module:f1")(
        lambda: {
            "tools": {"example": {"$id": "example1"}},
        }
    )
    fake_eps(name="aaa", value="test_module:f2")(
        lambda: {
            "tools": {"example": {"$id": "example2"}, "other": {"$id": "example3"}}
        }
    )
    for eps in (fake_eps, fake_eps.reverse()):
        monkeypatch.setattr(plugins, "iterate_entry_points", eps.get)
        # entry-point names closer to "zzzzzzzz..." have priority
        (plugin1, plugin2) = plugins.list_from_entry_points()
        print(plugin1, plugin2)
        assert plugin1.schema["$id"] == "example1"
        assert plugin2.schema["$id"] == "example3"


def test_custom_priority(monkeypatch):
    fake_eps = _FakeEntryPoints(monkeypatch)
    tool_eps = fake_eps.group("validate_pyproject.tool_schema")
    multi_eps = fake_eps.group("validate_pyproject.multi_schema")

    multi_schema = {"tools": {"example": {"$id": "multi-eps"}}}
    multi_eps(name="example", value="test_module:f")(lambda: multi_schema)

    @tool_eps(name="example", value="test_module1:f1")
    def tool_schema1(_name):
        return {"$id": "tool-eps-1"}

    @tool_eps(name="example", value="test_module2:f1")
    def tool_schema2(_name):
        return {"$id": "tool-eps-2"}

    monkeypatch.setattr(plugins, "iterate_entry_points", fake_eps.get)
    (winner,) = plugins.list_from_entry_points()  # default: tool with "higher" ep name
    assert winner.schema["$id"] == "tool-eps-2"

    tool_schema1.priority = 1.1
    (winner,) = plugins.list_from_entry_points()  # default: tool has priority
    assert winner.schema["$id"] == "tool-eps-1"

    tool_schema1.priority = 0.1
    tool_schema2.priority = 0.2
    multi_schema["priority"] = 0.9
    (winner,) = plugins.list_from_entry_points()  # custom higher priority wins
    assert winner.schema["$id"] == "multi-eps"


def test_broken_multi_plugin(monkeypatch):
    fake_eps = _FakeEntryPoints(monkeypatch, "validate_pyproject.multi_schema")
    fake_eps(name="broken", value="test_module.f")(lambda: {}["no-such-key"])
    monkeypatch.setattr(plugins, "iterate_entry_points", fake_eps.get)
    with pytest.raises(ErrorLoadingPlugin):
        plugins.list_from_entry_points()



import builtins
import importlib
import re
import shutil
import subprocess
import sys
from inspect import cleandoc
from pathlib import Path

import pytest
from fastjsonschema import JsonSchemaValueException

from validate_pyproject import _tomllib as tomllib
from validate_pyproject.pre_compile import cli, pre_compile

from .helpers import error_file, get_tools, get_tools_as_args

MAIN_FILE = "hello_world.py"  # Let's use something different than `__init__.py`


def _pre_compile_checks(path: Path):
    assert (path / "__init__.py").exists()
    assert (path / "__init__.py").read_text() == ""
    assert (path / MAIN_FILE).exists()
    files = [
        (MAIN_FILE, "def validate("),
        (MAIN_FILE, "from .error_reporting import detailed_errors, ValidationError"),
        ("error_reporting.py", "def detailed_errors("),
        ("fastjsonschema_exceptions.py", "class JsonSchemaValueException"),
        ("fastjsonschema_validations.py", "def validate("),
        ("extra_validations.py", "def validate"),
        ("formats.py", "def pep508("),
        ("NOTICE", "The relevant copyright notes and licenses are included below"),
    ]
    for file, content in files:
        assert (path / file).exists()
        assert content in (path / file).read_text()

    # Make sure standard replacements work
    for file in ("fastjsonschema_validations.py", "error_reporting.py"):
        file_contents = (path / file).read_text()
        assert "from fastjsonschema" not in file_contents
        assert "from ._vendor.fastjsonschema" not in file_contents
        assert "from validate_pyproject._vendor.fastjsonschema" not in file_contents
        assert "from .fastjsonschema_exceptions" in file_contents

    # Make sure the pre-compiled lib works
    script = f"""
    from {path.stem} import {Path(MAIN_FILE).stem} as mod

    assert issubclass(mod.ValidationError, mod.JsonSchemaValueException)

    example = {{
        "project": {{"name": "proj", "version": 42}}
    }}
    assert mod.validate(example) == example
    """
    cmd = [sys.executable, "-c", cleandoc(script)]
    error = r".project\.version. must be string"
    with pytest.raises(subprocess.CalledProcessError) as exc_info:
        subprocess.check_output(cmd, cwd=path.parent, stderr=subprocess.STDOUT)

    assert re.search(error, str(exc_info.value.output, "utf-8"))


def test_pre_compile_api(tmp_path):
    path = Path(tmp_path)
    pre_compile(path, MAIN_FILE)
    _pre_compile_checks(path)
    # Let's make sure it also works for __init__
    shutil.rmtree(str(path), ignore_errors=True)
    replacements = {"from fastjsonschema import": "from _vend.fastjsonschema import"}
    pre_compile(path, text_replacements=replacements)
    assert "def validate(" in (path / "__init__.py").read_text()
    assert not (path / MAIN_FILE).exists()
    file_contents = (path / "fastjsonschema_validations.py").read_text()
    assert "from _vend" in file_contents
    assert "from fastjsonschema" not in file_contents


def test_vendoring_cli(tmp_path):
    path = Path(tmp_path)
    cli.run(["-O", str(path), "-M", MAIN_FILE])
    _pre_compile_checks(Path(path))
    # Let's also try to test JSON replacements
    shutil.rmtree(str(path), ignore_errors=True)
    replacements = '{"from fastjsonschema import": "from _vend.fastjsonschema import"}'
    cli.run(["-O", str(path), "-R", replacements])
    file_contents = (path / "fastjsonschema_validations.py").read_text()
    assert "from _vend" in file_contents
    assert "from fastjsonschema" not in file_contents


# ---- Examples ----


PRE_COMPILED_NAME = "_validation"


def api_pre_compile(tmp_path, *, example: Path) -> Path:
    plugins = get_tools(example)
    return pre_compile(Path(tmp_path / PRE_COMPILED_NAME), extra_plugins=plugins)


def cli_pre_compile(tmp_path, *, example: Path) -> Path:
    args = get_tools_as_args(example)
    path = Path(tmp_path / PRE_COMPILED_NAME)
    cli.run([*args, "-O", str(path)])
    return path


_PRE_COMPILED = (api_pre_compile, cli_pre_compile)


@pytest.fixture
def pre_compiled_validate(monkeypatch):
    def _validate(vendored_path, toml_equivalent):
        assert PRE_COMPILED_NAME not in sys.modules
        importlib.invalidate_caches()
        with monkeypatch.context() as m:
            # Make sure original imports are not used
            _disable_import(m, "fastjsonschema")
            _disable_import(m, "validate_pyproject")
            # Make newly generated package available for importing
            m.syspath_prepend(str(vendored_path.parent))
            mod = __import__(PRE_COMPILED_NAME)
            print(list(vendored_path.glob("*")))
            print(mod, "\n\n", dir(mod))
            try:
                return mod.validate(toml_equivalent)
            except mod.JsonSchemaValueException as ex:
                # Let's translate the exceptions so we have identical classes
                new_ex = JsonSchemaValueException(
                    ex.message, ex.value, ex.name, ex.definition, ex.rule
                )
                raise new_ex from ex
            finally:
                all_modules = [
                    mod
                    for mod in sys.modules
                    if mod.startswith(f"{PRE_COMPILED_NAME}.")
                ]
                for mod in all_modules:
                    del sys.modules[mod]
                del sys.modules[PRE_COMPILED_NAME]

    return _validate


@pytest.mark.parametrize("pre_compiled", _PRE_COMPILED)
def test_examples_api(tmp_path, pre_compiled_validate, example, pre_compiled):
    toml_equivalent = tomllib.loads(example.read_text())
    pre_compiled_path = pre_compiled(Path(tmp_path), example=example)
    assert pre_compiled_validate(pre_compiled_path, toml_equivalent) is not None


@pytest.mark.parametrize("pre_compiled", _PRE_COMPILED)
def test_invalid_examples_api(
    tmp_path, pre_compiled_validate, invalid_example, pre_compiled
):
    expected_error = error_file(invalid_example).read_text("utf-8")
    toml_equivalent = tomllib.loads(invalid_example.read_text())
    pre_compiled_path = pre_compiled(Path(tmp_path), example=invalid_example)
    with pytest.raises(JsonSchemaValueException) as exc_info:
        pre_compiled_validate(pre_compiled_path, toml_equivalent)
    exception_message = str(exc_info.value)
    print("rule", "=", exc_info.value.rule)
    print("rule_definition", "=", exc_info.value.rule_definition)
    print("definition", "=", exc_info.value.definition)
    for error in expected_error.splitlines():
        assert error in exception_message


def _disable_import(monkeypatch, name):
    orig = builtins.__import__

    def _import(import_name, *args, **kwargs):
        if import_name == name or import_name.startswith(f"{name}."):
            raise ImportError(name)
        return orig(import_name, *args, **kwargs)

    monkeypatch.setattr(builtins, "__import__", _import)



import io
import os
from unittest.mock import Mock

import pytest

from validate_pyproject import caching, http, remote


@pytest.fixture(autouse=True)
def no_cache_env_var(monkeypatch):
    monkeypatch.delenv("VALIDATE_PYPROJECT_CACHE_REMOTE", raising=False)


def fn1(arg: str) -> io.StringIO:
    return io.StringIO("42")


def fn2(arg: str) -> io.StringIO:
    raise RuntimeError("should not be called")


def test_as_file(tmp_path):
    # The first call should create a file and return its contents
    cache_path = caching.path_for("hello-world", tmp_path)
    assert not cache_path.exists()

    with caching.as_file(fn1, "hello-world", tmp_path) as f:
        assert f.read() == b"42"

    assert cache_path.exists()
    assert cache_path.read_text("utf-8") == "42"

    # Any further calls using the same ``arg`` should reuse the file
    # and NOT call the function
    with caching.as_file(fn2, "hello-world", tmp_path) as f:
        assert f.read() == b"42"

    # If the file is deleted, then the function should be called
    cache_path.unlink()
    with pytest.raises(RuntimeError, match="should not be called"):
        caching.as_file(fn2, "hello-world", tmp_path)


def test_as_file_no_cache():
    # If no cache directory is passed, the orig function should
    # be called straight away:
    with pytest.raises(RuntimeError, match="should not be called"):
        caching.as_file(fn2, "hello-world")


def test_path_for_no_cache(monkeypatch):
    cache_path = caching.path_for("hello-world", None)
    assert cache_path is None


@pytest.mark.uses_network
@pytest.mark.skipif(
    os.getenv("VALIDATE_PYPROJECT_NO_NETWORK") or os.getenv("NO_NETWORK"),
    reason="Disable tests that depend on network",
)
class TestIntegration:
    def test_cache_open_url(self, tmp_path, monkeypatch):
        open_url = Mock(wraps=http.open_url)
        monkeypatch.setattr(http, "open_url", open_url)

        # The first time it is called, it will cache the results into a file
        url = (
            "https://raw.githubusercontent.com/abravalheri/validate-pyproject/main/"
            "src/validate_pyproject/pyproject_toml.schema.json"
        )
        cache_path = caching.path_for(url, tmp_path)
        assert not cache_path.exists()

        with caching.as_file(http.open_url, url, tmp_path) as f:
            assert b"build-system" in f.read()

        open_url.assert_called_once()
        assert cache_path.exists()
        assert "build-system" in cache_path.read_text("utf-8")

        # The second time, it will not reach the network, and use the file contents
        open_url.reset_mock()
        _, contents = remote.load_from_uri(url, cache_dir=tmp_path)

        assert "build-system" in contents["properties"]
        open_url.assert_not_called()



from collections.abc import Mapping
from functools import partial, wraps

import fastjsonschema as FJS
import pytest

from validate_pyproject import _tomllib as tomllib
from validate_pyproject import api, errors, plugins, types

PYPA_SPECS = "https://packaging.python.org/en/latest/specifications"


def test_load():
    spec = api.load("pyproject_toml")
    assert isinstance(spec, Mapping)

    assert spec["$id"] == f"{PYPA_SPECS}/declaring-build-dependencies/"

    spec = api.load("project_metadata")
    assert spec["$id"] == f"{PYPA_SPECS}/pyproject-toml/"


def test_load_plugin():
    spec = api.load_builtin_plugin("distutils")
    assert spec["$id"].startswith("https://setuptools.pypa.io")
    assert "deprecated/distutils" in spec["$id"]

    spec = api.load_builtin_plugin("setuptools")
    assert spec["$id"].startswith("https://setuptools.pypa.io")
    assert "pyproject" in spec["$id"]


class TestRegistry:
    def test_with_plugins(self):
        plg = plugins.list_from_entry_points()
        registry = api.SchemaRegistry(plg)
        main_schema = registry[registry.main]
        project = main_schema["properties"]["project"]
        assert project["$ref"] == f"{PYPA_SPECS}/pyproject-toml/"
        tool = main_schema["properties"]["tool"]
        assert "setuptools" in tool["properties"]
        assert "$ref" in tool["properties"]["setuptools"]

    def fake_plugin(self, name, schema_version=7, end="#"):
        schema = {
            "$id": f"https://example.com/{name}.schema.json",
            "$schema": f"http://json-schema.org/draft-{schema_version:02d}/schema{end}",
            "type": "object",
        }
        return types.Schema(schema)

    @pytest.mark.parametrize("end", ["", "#"], ids=["no#", "with#"])
    def test_schema_ending(self, end):
        fn = wraps(self.fake_plugin)(partial(self.fake_plugin, end=end))
        plg = plugins.PluginWrapper("plugin", fn)
        registry = api.SchemaRegistry([plg])
        main_schema = registry[registry.main]
        assert main_schema["$schema"] == "http://json-schema.org/draft-07/schema#"

    def test_incompatible_versions(self):
        fn = wraps(self.fake_plugin)(partial(self.fake_plugin, schema_version=8))
        plg = plugins.PluginWrapper("plugin", fn)
        with pytest.raises(errors.InvalidSchemaVersion):
            api.SchemaRegistry([plg])

    def test_duplicated_id_different_tools(self):
        schema = self.fake_plugin("plg")
        fn = wraps(self.fake_plugin)(lambda _: schema)  # Same ID
        plg = [plugins.PluginWrapper(f"plg{i}", fn) for i in range(2)]
        with pytest.raises(errors.SchemaWithDuplicatedId):
            api.SchemaRegistry(plg)

    def test_allow_overwrite_same_tool(self):
        plg = [plugins.PluginWrapper("plg", self.fake_plugin) for _ in range(2)]
        registry = api.SchemaRegistry(plg)
        sid = self.fake_plugin("plg")["$id"]
        assert sid in registry

    def test_missing_id(self):
        def _fake_plugin(name):
            plg = dict(self.fake_plugin(name))
            del plg["$id"]
            return types.Schema(plg)

        plg = plugins.PluginWrapper("plugin", _fake_plugin)
        with pytest.raises(errors.SchemaMissingId):
            api.SchemaRegistry([plg])


class TestValidator:
    example_toml = """\
    [project]
    name = "myproj"
    version = "0"

    [tool.setuptools]
    zip-safe = false
    packages = {find = {}}
    """

    @property
    def valid_example(self):
        return tomllib.loads(self.example_toml)

    @property
    def invalid_example(self):
        example = self.valid_example
        example["tool"]["setuptools"]["zip-safe"] = {"hello": "world"}
        return example

    def test_valid(self):
        validator = api.Validator()
        assert validator(self.valid_example) is not None

    def test_invalid(self):
        validator = api.Validator()
        with pytest.raises(FJS.JsonSchemaValueException):
            validator(self.invalid_example)

    # ---

    def plugin(self, tool):
        plg = plugins.list_from_entry_points(filtering=lambda e: e.name == tool)
        return plg[0]

    TOOLS = ("distutils", "setuptools")

    @pytest.mark.parametrize("tool, other_tool", zip(TOOLS, reversed(TOOLS)))
    def test_plugin_not_enabled(self, tool, other_tool):
        plg = self.plugin(tool)
        validator = api.Validator([plg])
        registry = validator.registry
        main_schema = registry[registry.main]
        assert tool in main_schema["properties"]["tool"]["properties"]
        assert other_tool not in main_schema["properties"]["tool"]["properties"]
        tool_properties = main_schema["properties"]["tool"]["properties"]
        assert tool_properties[tool]["$ref"] == plg.schema["$id"]

    def test_invalid_but_plugin_not_enabled(self):
        # When the plugin is not enabled, the validator should ignore the tool
        validator = api.Validator([self.plugin("distutils")])
        try:
            assert validator(self.invalid_example) is not None
        except Exception:
            registry = validator.registry
            main_schema = registry[registry.main]
            assert "setuptools" not in main_schema["properties"]["tool"]["properties"]
            import json

            assert "setuptools" not in json.dumps(main_schema)
            raise



"""``sphinx-apidoc`` only allows users to specify "exclude patterns" but not
"include patterns". This module solves that gap.
"""

import shutil
from pathlib import Path

MODULE_TEMPLATE = """
``{name}``
~~{underline}~~

.. automodule:: {name}
   :members:{_members}
   :undoc-members:
   :show-inheritance:
   :special-members: __call__
"""

__location__ = Path(__file__).parent


def gen_stubs(module_dir: str, output_dir: str):
    shutil.rmtree(output_dir, ignore_errors=True)  # Always start fresh
    out = Path(output_dir)
    out.mkdir(parents=True, exist_ok=True)
    manifest = shutil.copy(__location__ / "modules.rst.in", out / "modules.rst")
    for module in iter_public(manifest):
        text = module_template(module)
        Path(output_dir, f"{module}.rst").write_text(text, encoding="utf-8")


def iter_public(manifest):
    toc = Path(manifest).read_text(encoding="utf-8")
    lines = (x.strip() for x in toc.splitlines())
    return (x for x in lines if x.startswith("validate_pyproject."))


def module_template(name: str, *members: str) -> str:
    underline = "~" * len(name)
    _members = (" " + ", ".join(members)) if members else ""
    return MODULE_TEMPLATE.format(name=name, underline=underline, _members=_members)



import traceback

import pytest

from validate_pyproject.pre_compile import cli


@pytest.mark.parametrize("replacements", ['["a", "b"]', "{invalid: json}"])
def test_invalid_replacements(tmp_path, replacements):
    with pytest.raises(SystemExit) as exc:
        cli.run(["-O", str(tmp_path), "-R", replacements])

    e = exc.value
    trace = "".join(traceback.format_exception(e.__class__, e, e.__traceback__))
    assert "--replacements: invalid" in trace
    assert replacements in trace



"""Test summary generation from schema examples"""

import json
from pathlib import Path

import pytest

from validate_pyproject.error_reporting import _SummaryWriter

EXAMPLE_FOLDER = Path(__file__).parent / "json_schema_summary"
EXAMPLES = (p.name for p in EXAMPLE_FOLDER.glob("*"))


def load_example(file):
    text = file.read_text(encoding="utf-8")
    schema, _, summary = text.partition("# - # - # - #\n")

    # # Auto fix examples:
    # fixed = _SummaryWriter()(json.loads(schema))
    # file.write_text(text.replace(summary, fixed), encoding="utf-8")

    return json.loads(schema), summary


@pytest.mark.parametrize("example", EXAMPLES)
def test_summary_generation(example):
    schema, expected = load_example(EXAMPLE_FOLDER / example)
    summarize = _SummaryWriter()
    summary = summarize(schema)
    assert summary == expected



from typing import Any, Dict

import fastjsonschema

from . import api, plugins

__all__ = ["VPP001", "repo_review_checks", "repo_review_families"]


class VPP001:
    """Validate pyproject.toml"""

    family = "validate-pyproject"

    @staticmethod
    def check(pyproject: Dict[str, Any]) -> str:
        validator = api.Validator()
        try:
            validator(pyproject)
            return ""
        except fastjsonschema.JsonSchemaValueException as e:
            return f"Invalid pyproject.toml! Error: {e}"


def repo_review_checks() -> Dict[str, VPP001]:
    return {"VPP001": VPP001()}


def repo_review_families(pyproject: Dict[str, Any]) -> Dict[str, Dict[str, str]]:
    has_distutils = "distutils" in pyproject.get("tool", {})
    plugin_list = plugins.list_from_entry_points(
        lambda e: e.name != "distutils" or has_distutils
    )
    plugin_names = (f"`[tool.{n.tool}]`" for n in plugin_list if n.tool)
    descr = f"Checks `[build-system]`, `[project]`, {', '.join(plugin_names)}"
    return {"validate-pyproject": {"name": "Validate-PyProject", "description": descr}}



import copy
import logging
from pathlib import Path

import pytest

from validate_pyproject import _tomllib as tomllib
from validate_pyproject import api, cli
from validate_pyproject.error_reporting import ValidationError

from .helpers import error_file, get_tools, get_tools_as_args


def test_examples_api(example: Path) -> None:
    load_tools = get_tools(example)

    toml_equivalent = tomllib.loads(example.read_text())
    copy_toml = copy.deepcopy(toml_equivalent)
    validator = api.Validator(extra_plugins=load_tools)
    assert validator(toml_equivalent) is not None
    assert toml_equivalent == copy_toml


def test_examples_cli(example: Path) -> None:
    args = get_tools_as_args(example)

    assert cli.run(["--dump-json", str(example), *args]) == 0  # no errors


def test_invalid_examples_api(invalid_example: Path) -> None:
    load_tools = get_tools(invalid_example)

    expected_error = error_file(invalid_example).read_text("utf-8")
    toml_equivalent = tomllib.loads(invalid_example.read_text())
    validator = api.Validator(extra_plugins=load_tools)
    with pytest.raises(ValidationError) as exc_info:
        validator(toml_equivalent)
    exception_message = str(exc_info.value)
    summary = exc_info.value.summary
    for error in expected_error.splitlines():
        assert error in exception_message
        assert error in summary


def test_invalid_examples_cli(invalid_example: Path, caplog) -> None:
    args = get_tools_as_args(invalid_example)

    caplog.set_level(logging.DEBUG)
    expected_error = error_file(invalid_example).read_text("utf-8")
    with pytest.raises(SystemExit) as exc_info:
        cli.main([str(invalid_example), *args])
    assert exc_info.value.args == (1,)
    for error in expected_error.splitlines():
        assert error in caplog.text



from typing import Callable, Mapping, NewType, TypeVar

T = TypeVar("T", bound=Mapping)

Schema = NewType("Schema", Mapping)
"""JSON Schema represented as a Python dict"""

ValidationFn = Callable[[T], T]
"""Custom validation function.
It should receive as input a mapping corresponding to the whole
``pyproject.toml`` file and raise a :exc:`fastjsonschema.JsonSchemaValueException`
if it is not valid.
"""

FormatValidationFn = Callable[[str], bool]
"""Should return ``True`` when the input string satisfies the format"""

Plugin = Callable[[str], Schema]
"""A plugin is something that receives the name of a `tool` sub-table
(as defined  in PEPPEP621) and returns a :obj:`Schema`.

For example ``plugin("setuptools")`` should return the JSON schema for the
``[tool.setuptools]`` table of a ``pyproject.toml`` file.
"""



from typing import Any

from ..pre_compile import cli
from . import _deprecated


def run(*args: Any, **kwargs: Any) -> Any:
    return _deprecated(run, cli.run)(*args, **kwargs)


def main(*args: Any, **kwargs: Any) -> Any:
    return _deprecated(run, cli.main)(*args, **kwargs)



from argparse import Namespace
from pathlib import Path

import pytest

from validate_pyproject import _tomllib as tomllib
from validate_pyproject.repo_review import repo_review_checks, repo_review_families

DIR = Path(__file__).parent.resolve()
EXAMPLES = DIR / "examples"
INVALID_EXAMPLES = DIR / "invalid-examples"


@pytest.fixture
def repo_review_processor():
    try:
        from repo_review import processor

        return processor
    except ImportError:

        class _Double:  # just for the sake of Python < 3.10 coverage
            @staticmethod
            def process(file: Path):
                pyproject = (file / "pyproject.toml").read_text(encoding="utf-8")
                opts = tomllib.loads(pyproject)
                checks = (
                    checker.check(opts) == ""  # No errors
                    for checker in repo_review_checks().values()
                )
                return Namespace(
                    families=repo_review_families(opts),
                    results=[Namespace(result=check) for check in checks],
                )

        return _Double


@pytest.mark.parametrize("name", ["atoml", "flit", "pdm", "pep_text", "trampolim"])
def test_valid_example(repo_review_processor, name: str) -> None:
    processed = repo_review_processor.process(EXAMPLES / name)
    assert all(r.result for r in processed.results), f"{processed.results}"


@pytest.mark.parametrize("name", ["pdm/invalid-version", "pdm/redefining-as-dynamic"])
def test_invalid_example(repo_review_processor, name: str) -> None:
    processed = repo_review_processor.process(INVALID_EXAMPLES / name)
    assert any(not r.result and r.result is not None for r in processed.results), (
        f"{processed.results}"
    )


def test_no_distutils(repo_review_processor) -> None:
    processed = repo_review_processor.process(EXAMPLES / "pep_text")
    family = processed.families["validate-pyproject"]
    assert "[tool.setuptools]" in family["description"]
    assert "[tool.distutils]" not in family["description"]


def test_has_distutils(repo_review_processor, tmp_path: Path) -> None:
    d = tmp_path / "no-distutils"
    d.mkdir()
    d.joinpath("pyproject.toml").write_text("[tool.distutils]")
    processed = repo_review_processor.process(d)
    family = processed.families["validate-pyproject"]
    assert "[tool.setuptools]" in family["description"]
    assert "[tool.distutils]" in family["description"]



import pytest

from validate_pyproject.vendoring import cli, vendorify


def test_api(tmp_path):
    with pytest.warns(DeprecationWarning, match="will be removed"):
        vendorify(tmp_path)


def test_cli(tmp_path):
    with pytest.warns(DeprecationWarning, match="will be removed"):
        cli.run(["-O", str(tmp_path)])


def test_main(tmp_path):
    with pytest.warns(DeprecationWarning, match="will be removed"):
        cli.main(["-O", str(tmp_path)])



#!/usr/bin/env python3

import argparse
import json


def convert_tree(tree: dict[str, object]) -> None:
    for key, value in list(tree.items()):
        match key, value:
            case "$$description", list():
                tree["description"] = " ".join(value)
                del tree["$$description"]
            case "$id", str():
                del tree["$id"]
            case _, dict():
                convert_tree(value)
            case _, list():
                for item in value:
                    if isinstance(item, dict):
                        convert_tree(item)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("schema", help="JSONSchema to convert")
    args = parser.parse_args()

    with open(args.schema, encoding="utf-8") as f:
        schema = json.load(f)

    convert_tree(schema)
    schema["$id"] = "https://json.schemastore.org/setuptools.json"

    print(json.dumps(schema, indent=2))



import inspect
import io
import logging
import sys
from pathlib import Path
from unittest.mock import Mock
from uuid import uuid4

import pytest
from fastjsonschema import JsonSchemaValueException

from validate_pyproject import cli, errors, plugins


class TestHelp:
    def test_list_default_plugins(self, capsys):
        with pytest.raises(SystemExit):
            cli.main(["--help"])
        captured = capsys.readouterr()
        assert "setuptools" in captured.out
        assert "distutils" in captured.out

    def test_no_plugins(self, capsys):
        with pytest.raises(SystemExit):
            cli.parse_args(["--help"], plugins=[])
        captured = capsys.readouterr()
        assert "setuptools" not in captured.out
        assert "distutils" not in captured.out

    def test_custom_plugins(self, capsys):
        fake_plugin = plugins.PluginWrapper("my42", lambda _: {})
        with pytest.raises(SystemExit):
            cli.parse_args(["--help"], plugins=[fake_plugin])
        captured = capsys.readouterr()
        assert "my42" in captured.out


def parse_args(args):
    plg = plugins.list_from_entry_points()
    return cli.parse_args(args, plg)


simple_example = """\
[project]
name = "myproj"
version = "0"

[tool.setuptools]
zip-safe = false
packages = {find = {}}
"""


def write_example(dir_path, *, name="pyproject.toml", _text=simple_example):
    path = Path(dir_path, name)
    path.write_text(_text, "UTF-8")
    return path


def write_invalid_example(dir_path, *, name="pyproject.toml"):
    text = simple_example.replace("zip-safe = false", "zip-safe = { hello = 'world' }")
    return write_example(dir_path, name=name, _text=text)


@pytest.fixture
def valid_example(tmp_path):
    return write_example(tmp_path)


@pytest.fixture
def invalid_example(tmp_path):
    return write_invalid_example(tmp_path)


class TestEnable:
    TOOLS = ("setuptools", "distutils")

    @pytest.mark.parametrize("tool", TOOLS)
    def test_parse(self, valid_example, tool):
        params = parse_args([str(valid_example), "-E", tool])
        assert len(params.plugins) == 1
        assert params.plugins[0].tool == tool

        # Meta test:
        schema = params.plugins[0].schema
        if tool == "setuptools":
            assert "zip-safe" in schema["properties"]
            assert schema["properties"]["zip-safe"]["type"] == "boolean"

    def test_valid(self, valid_example):
        assert cli.main([str(valid_example), "-E", "setuptools"]) == 0

    def test_invalid(self, invalid_example):
        print(invalid_example.read_text())
        with pytest.raises(JsonSchemaValueException):
            cli.run([str(invalid_example), "-E", "setuptools"])

    def test_invalid_not_enabled(self, invalid_example):
        # When the plugin is not enabled, the validator should ignore the tool
        assert cli.main([str(invalid_example), "-E", "distutils"]) == 0


class TestDisable:
    TOOLS = ("setuptools", "distutils")

    @pytest.mark.parametrize("tool, other_tool", zip(TOOLS, reversed(TOOLS)))
    def test_parse(self, valid_example, tool, other_tool):
        all_plugins = parse_args([str(valid_example), "-D", tool]).plugins
        our_plugins = [p for p in all_plugins if p.id.startswith("validate_pyproject")]
        assert len(our_plugins) == 1
        assert our_plugins[0].tool == other_tool

    def test_valid(self, valid_example):
        assert cli.run([str(valid_example), "-D", "distutils"]) == 0

    def test_invalid(self, invalid_example):
        print(invalid_example.read_text())
        with pytest.raises(JsonSchemaValueException):
            cli.run([str(invalid_example), "-D", "distutils"])

    def test_invalid_disabled(self, invalid_example):
        # When the plugin is disabled, the validator should ignore the tool
        assert cli.main([str(invalid_example), "-D", "setuptools"]) == 0


class TestInput:
    def test_inform_user_about_stdin(self, monkeypatch):
        print_mock = Mock()
        fake_stdin = io.StringIO('[project]\nname="test"\nversion="0.42"\n')
        with monkeypatch.context() as ctx:
            ctx.setattr("validate_pyproject.cli._STDIN", fake_stdin)
            ctx.setattr("sys.argv", ["validate-pyproject"])
            ctx.setattr("builtins.print", print_mock)
            cli.run()
        calls = print_mock.call_args_list
        assert any("input via `stdin`" in str(args[0]) for args, _kwargs in calls)


class TestOutput:
    def test_valid(self, capsys, valid_example):
        cli.main([str(valid_example)])
        captured = capsys.readouterr()
        assert "valid" in captured.out.lower()

    def test_invalid(self, caplog, invalid_example):
        caplog.set_level(logging.DEBUG)
        with pytest.raises(SystemExit):
            cli.main([str(invalid_example)])
        captured = caplog.text.lower()
        assert "`tool.setuptools.zip-safe` must be boolean" in captured
        assert "offending rule" in captured
        assert "given value" in captured
        assert '"type": "boolean"' in captured


def test_multiple_files(tmp_path, capsys):
    N = 3

    valid_files = [
        write_example(tmp_path, name=f"valid-pyproject{i}.toml") for i in range(N)
    ]
    cli.run(map(str, valid_files))
    captured = capsys.readouterr().out.lower()
    number_valid = captured.count("valid file:")
    assert number_valid == N

    invalid_files = [
        write_invalid_example(tmp_path, name=f"invalid-pyproject{i}.toml")
        for i in range(N + 3)
    ]
    with pytest.raises(SystemExit):
        cli.main(map(str, valid_files + invalid_files))

    repl = str(uuid4())
    captured = capsys.readouterr().out.lower()
    captured = captured.replace("invalid file:", repl)
    number_invalid = captured.count(repl)
    number_valid = captured.count("valid file:")
    captured = captured.replace(repl, "invalid file:")
    assert number_valid == N
    assert number_invalid == N + 3


def test_missing_toolname(tmp_path, capsys):
    example = write_example(tmp_path, name="valid-pyproject.toml")
    with pytest.raises(
        errors.URLMissingTool,
        match=r"Correct form is '--tool <tool-name>=http://json\.schemastore\.org/poetry\.toml', with an optional",
    ):
        cli.run(["--tool=http://json.schemastore.org/poetry.toml", str(example)])


def test_bad_url(tmp_path, capsys):
    example = write_example(tmp_path, name="valid-pyproject.toml")
    with pytest.raises(ValueError, match="URL must start with 'http:' or 'https:'"):
        cli.run(
            ["--tool", "poetry=file://json.schemastore.org/poetry.toml", str(example)]
        )


def test_bad_extra_url(tmp_path, capsys):
    example = write_example(tmp_path, name="valid-pyproject.toml")
    with pytest.raises(ValueError, match="URL must start with 'http:' or 'https:'"):
        cli.run(["--tool", "=file://json.schemastore.org/poetry.toml", str(example)])


@pytest.mark.skipif(sys.version_info[:2] < (3, 11), reason="requires 3.11+")
def test_parser_is_tomllib():
    """Make sure Python >= 3.11 uses tomllib instead of tomli"""
    module_name = inspect.getmodule(cli.tomllib.loads).__name__
    assert module_name.startswith("tomllib")



import logging
from inspect import cleandoc

import pytest
from fastjsonschema import validate

from validate_pyproject.api import FORMAT_FUNCTIONS
from validate_pyproject.error_reporting import ValidationError, detailed_errors

EXAMPLES = {
    "const": {
        "schema": {"const": 42},
        "value": 13,
        "message": "`data` must be 42",
        "debug_info": "**SKIP-TEST**",
    },
    "container": {
        "schema": {"type": "array", "contains": True},
        "value": [],
        "message": "`data` must not be empty",
        "debug_info": "**SKIP-TEST**",
    },
    "type": {
        "schema": {"anyOf": [{"not": {"type": ["string", "number"]}}]},
        "value": 42,
        "message": """
            `data` cannot be validated by any definition:

                - (*NOT* the following):
                    type: [string, number]
        """,
        "debug_info": "**SKIP-TEST**",
    },
    "oneOf": {
        "schema": {
            "oneOf": [{"type": "string", "format": "pep440"}, {"type": "integer"}]
        },
        "value": {"use_scm": True},
        "message": """
            `data` must be valid exactly by one definition (0 matches found):

                - {type: string, format: 'pep440'}
                - {type: integer}
        """,
        "debug_info": """
            GIVEN VALUE:
                {
                    "use_scm": true
                }

            OFFENDING RULE: 'oneOf'

            DEFINITION:
                {
                    "oneOf": [
                        {
                            "type": "string",
                            "format": "pep440"
                        },
                        {
                            "type": "integer"
                        }
                    ]
                }

            For more details about `format` see
        """,
    },
    "description": {
        "schema": {"type": "string", "description": "Lorem ipsum dolor sit amet"},
        "value": {"name": 42},
        "message": "`data` must be string",
        "debug_info": """
            DESCRIPTION:
                Lorem ipsum dolor sit amet
        """,
    },
    "$$description": {
        "schema": {
            "properties": {
                "name": {
                    "type": "string",
                    "$$description": [
                        "Lorem ipsum dolor sit amet, consectetur adipiscing elit,",
                        "sed do eiusmod tempor incididunt ut labore et dolore magna",
                        "aliqua. Ut enim ad minim veniam, quis nostrud exercitation",
                        "ullamco laboris nisi ut aliquip ex ea commodo consequat.",
                    ],
                }
            }
        },
        "value": {"name": 42},
        "message": "`name` must be string",
        "debug_info": """
            DESCRIPTION:
                Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod
                tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam,
                quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo
                consequat.

            GIVEN VALUE:
                42

            OFFENDING RULE: 'type'

            DEFINITION:
                {
                    "type": "string"
                }
        """,
    },
}


@pytest.mark.parametrize("example", EXAMPLES.keys())
def test_error_reporting(caplog, example):
    schema = EXAMPLES[example]["schema"]
    value = EXAMPLES[example]["value"]
    message = cleandoc(EXAMPLES[example]["message"])
    debug_info = cleandoc(EXAMPLES[example]["debug_info"])

    try:
        with caplog.at_level(logging.CRITICAL), detailed_errors():
            validate(schema, value, formats=FORMAT_FUNCTIONS)
    except ValidationError as ex:
        assert ex.message.strip() == message
        assert ex.message == ex.summary
        assert "GIVEN VALUE:" in ex.details
        assert "DEFINITION:" in ex.details

    try:
        with caplog.at_level(logging.DEBUG), detailed_errors():
            validate(schema, value, formats=FORMAT_FUNCTIONS)
    except ValidationError as ex:
        assert "GIVEN VALUE:" in ex.message
        assert "DEFINITION:" in ex.message
        assert ex.summary in ex.message
        if debug_info != "**SKIP-TEST**":
            assert debug_info in ex.details



# This module is intentionally kept minimal,
# so that it can be imported without triggering imports outside stdlib.
import io
import sys
from urllib.request import urlopen

if sys.platform == "emscripten" and "pyodide" in sys.modules:
    from pyodide.http import open_url
else:

    def open_url(url: str) -> io.StringIO:
        if not url.startswith(("http:", "https:")):
            raise ValueError("URL must start with 'http:' or 'https:'")
        with urlopen(url) as response:  # noqa: S310
            return io.StringIO(response.read().decode("utf-8"))



import json
import logging
import typing
import urllib.parse
from typing import Generator, Optional, Tuple

from . import caching, errors, http
from .types import Schema

if typing.TYPE_CHECKING:
    import sys

    if sys.version_info < (3, 11):
        from typing_extensions import Self
    else:
        from typing import Self


__all__ = ["RemotePlugin", "load_store"]


_logger = logging.getLogger(__name__)


def load_from_uri(
    tool_uri: str, cache_dir: Optional[caching.PathLike] = None
) -> Tuple[str, Schema]:
    tool_info = urllib.parse.urlparse(tool_uri)
    if tool_info.netloc:
        url = f"{tool_info.scheme}://{tool_info.netloc}{tool_info.path}"
        download = caching.as_file(http.open_url, url, cache_dir)
        with download as f:
            contents = json.load(f)
    else:
        with open(tool_info.path, "rb") as f:
            contents = json.load(f)
    return tool_info.fragment, contents


class RemotePlugin:
    def __init__(self, *, tool: str, schema: Schema, fragment: str = ""):
        self.tool = tool
        self.schema = schema
        self.fragment = fragment
        self.id = self.schema["$id"]
        self.help_text = f"{tool} <external>"

    @classmethod
    def from_url(cls, tool: str, url: str) -> "Self":
        fragment, schema = load_from_uri(url)
        return cls(tool=tool, schema=schema, fragment=fragment)

    @classmethod
    def from_str(cls, tool_url: str) -> "Self":
        tool, _, url = tool_url.partition("=")
        if not url:
            raise errors.URLMissingTool(tool)
        return cls.from_url(tool, url)


def load_store(pyproject_url: str) -> Generator[RemotePlugin, None, None]:
    """
    Takes a URL / Path and loads the tool table, assuming it is a set of ref's.
    Currently ignores "inline" sections. This is the format that SchemaStore
    (https://json.schemastore.org/pyproject.json) is in.
    """

    fragment, contents = load_from_uri(pyproject_url)
    if fragment:
        _logger.error(
            f"Must not be called with a fragment, got {fragment!r}"
        )  # pragma: no cover
    table = contents["properties"]["tool"]["properties"]
    for tool, info in table.items():
        if tool in {"setuptools", "distutils"}:
            pass  # built-in
        elif "$ref" in info:
            _logger.info(f"Loading {tool} from store: {pyproject_url}")
            rp = RemotePlugin.from_url(tool, info["$ref"])
            yield rp
            for values in rp.schema["properties"].values():
                url = values.get("$ref", "")
                if url.startswith(("https://", "https://")):
                    yield RemotePlugin.from_url("", url)
        else:
            _logger.warning(f"{tool!r} does not contain $ref")  # pragma: no cover


if typing.TYPE_CHECKING:
    from .plugins import PluginProtocol

    _: PluginProtocol = typing.cast("RemotePlugin", None)



import functools
import json
from pathlib import Path
from typing import Dict, List, Union

from validate_pyproject.remote import RemotePlugin, load_store

HERE = Path(__file__).parent.resolve()


def error_file(p: Path) -> Path:
    try:
        files = (p.with_name("errors.txt"), p.with_suffix(".errors.txt"))
        return next(f for f in files if f.exists())
    except StopIteration:
        raise FileNotFoundError(f"No error file found for {p}") from None


def get_test_config(example: Path) -> Dict[str, Union[str, Dict[str, str]]]:
    test_config = example.with_name("test_config.json")
    if test_config.is_file():
        with test_config.open(encoding="utf-8") as f:
            return json.load(f)
    return {}


@functools.lru_cache(maxsize=None)
def get_tools(example: Path) -> List[RemotePlugin]:
    config = get_test_config(example)
    tools: Dict[str, str] = config.get("tools", {})
    load_tools = [RemotePlugin.from_url(k, v) for k, v in tools.items()]
    store: str = config.get("store", "")
    if store:
        load_tools.extend(load_store(store))
    return load_tools


@functools.lru_cache(maxsize=None)
def get_tools_as_args(example: Path) -> List[str]:
    config = get_test_config(example)
    tools: Dict[str, str] = config.get("tools", {})
    load_tools = [f"--tool={k}={v}" for k, v in tools.items()]
    store: str = config.get("store", "")
    if store:
        load_tools.append(f"--store={store}")
    return load_tools
